@@ -24,6 +24,7 @@ Built-in Log General-Purpose Quantities
 .. autoclass:: InitTime
 .. autoclass:: CPUTime
 .. autoclass:: ETA
+.. autoclass:: MemoryHwm
 .. autofunction:: add_general_quantities
 
 Built-in Log Simulation-Related Quantities
@@ -64,9 +65,13 @@ __version__ = logpyle.version.VERSION_TEXT
 import logging
 logger = logging.getLogger(__name__)
 
-from typing import List, Callable, Union, Tuple, Optional, Dict
+from typing import (List, Callable, Union, Tuple, Optional, Dict, Any,
+                    TYPE_CHECKING, Iterable)
 from pytools.datatable import DataTable
 
+if TYPE_CHECKING:
+    import mpi4py
+
 
 # {{{ timing function
 
@@ -101,7 +106,8 @@ class LogQuantity:
 
     sort_weight = 0
 
-    def __init__(self, name: str, unit: str = None, description: str = None) -> None:
+    def __init__(self, name: str, unit: Optional[str] = None,
+                 description: Optional[str] = None) -> None:
         """Create a new quantity.
 
         Parameters
@@ -128,7 +134,7 @@ class LogQuantity:
         """Perform updates required at every :class:`LogManager` tick."""
         pass
 
-    def __call__(self) -> Optional[float]:
+    def __call__(self) -> Any:
         """Return the current value of the diagnostic represented by this
         :class:`LogQuantity` or None if no value is available.
 
@@ -156,7 +162,8 @@ class PostLogQuantity(LogQuantity):
 
 
 class MultiLogQuantity:
-    """A source of a list of loggable scalars gathered at the start of each time step.
+    """A source of a list of loggable scalars gathered at the start of each time
+    step.
 
     Quantity values are gathered in :meth:`LogManager.tick_before`.
 
@@ -167,8 +174,8 @@ class MultiLogQuantity:
     """
     sort_weight = 0
 
-    def __init__(self, names: List[str], units: List[str] = None,
-                 descriptions: List[str] = None) -> None:
+    def __init__(self, names: List[str], units: Optional[List[str]] = None,
+                 descriptions: Optional[List[str]] = None) -> None:
         """Create a new quantity.
 
         Parameters
@@ -185,11 +192,11 @@ class MultiLogQuantity:
         self.names = names
 
         if units is None:
-            units = len(names) * [None]  # type: ignore
+            units = len(names) * [""]
         self.units = units
 
         if descriptions is None:
-            descriptions = len(names) * [None]  # type: ignore
+            descriptions = len(names) * [""]
         self.descriptions = descriptions
 
     @property
@@ -201,7 +208,7 @@ class MultiLogQuantity:
         """Perform updates required at every :class:`LogManager` tick."""
         pass
 
-    def __call__(self) -> None:
+    def __call__(self) -> Iterable[Optional[float]]:
         """Return an iterable of the current values of the diagnostic represented
         by this :class:`MultiLogQuantity`.
 
@@ -244,14 +251,15 @@ class TimeTracker(DtConsumer):
 class SimulationLogQuantity(PostLogQuantity, DtConsumer):
     """A source of loggable scalars that needs to know the simulation timestep."""
 
-    def __init__(self, dt: Optional[float], name: str, unit: str = None,
-                 description: str = None) -> None:
+    def __init__(self, dt: Optional[float], name: str, unit: Optional[str] = None,
+                 description: Optional[str] = None) -> None:
         PostLogQuantity.__init__(self, name, unit, description)
         DtConsumer.__init__(self, dt)
 
 
 class PushLogQuantity(LogQuantity):
-    def __init__(self, name: str, unit: str = None, description: str = None) -> None:
+    def __init__(self, name: str, unit: Optional[str] = None,
+                 description: Optional[str] = None) -> None:
         LogQuantity.__init__(self, name, unit, description)
         self.value = None
 
@@ -268,8 +276,8 @@ class PushLogQuantity(LogQuantity):
 
 class CallableLogQuantityAdapter(LogQuantity):
     """Adapt a 0-ary callable as a :class:`LogQuantity`."""
-    def __init__(self, callable: Callable, name: str, unit: str = None,
-                 description: str = None) -> None:
+    def __init__(self, callable: Callable, name: str, unit: Optional[str] = None,
+                 description: Optional[str] = None) -> None:
         self.callable = callable
         LogQuantity.__init__(self, name, unit, description)
 
@@ -345,7 +353,7 @@ def _get_unique_id() -> str:
         from random import Random
         rng = Random()
         rng.seed()
-        for i in range(20):
+        for _ in range(20):
             checksum.update(str(rng.randrange(1 << 30)).encode("utf-32"))
         return checksum.hexdigest()
     else:
@@ -407,7 +415,7 @@ class LogManager:
     :meth:`tick_after` calls captures data for a single time state,
     namely that in which the data may have been *before* the "compute"
     step. However, some data (such as the length of the timestep taken
-    in a time-adpative method) may only be available *after* the completion
+    in a time-adaptive method) may only be available *after* the completion
     of the "compute..." stage, which is why :meth:`tick_after` exists.
 
     A :class:`LogManager` logs any number of named time series of floats to
@@ -435,6 +443,7 @@ class LogManager:
 
     .. automethod:: capture_warnings
     .. automethod:: add_watches
+    .. automethod:: set_watch_interval
     .. automethod:: set_constant
     .. automethod:: add_quantity
 
@@ -444,8 +453,10 @@ class LogManager:
     .. automethod:: tick_after
     """
 
-    def __init__(self, filename: str = None, mode: str = "r", mpi_comm=None,
-                 capture_warnings: bool = True, commit_interval: float = 90) -> None:
+    def __init__(self, filename: Optional[str] = None, mode: str = "r",
+                 mpi_comm: Optional["mpi4py.MPI.Comm"] = None,
+                 capture_warnings: bool = True, commit_interval: float = 90,
+                 watch_interval: float = 1.0) -> None:
         """Initialize this log manager instance.
 
         :param filename: If given, the filename to which this log is bound.
@@ -454,13 +465,14 @@ class LogManager:
           database is initially empty. May also be "wu" to indicate that
           a unique filename should be chosen automatically. May also be "wo"
           to indicate that the file should be overwritten.
-        :arg mpi_comm: A `mpi4py.MPI.Comm`. If given, logs are
-            periodically synchronized to the head node, which then writes them
-            out to disk.
+        :arg mpi_comm: An optional :class:`mpi4py.MPI.Comm` object.
+          If given, logs are periodically synchronized to the head node,
+          which then writes them out to disk.
         :param capture_warnings: Tap the Python warnings facility and save warnings
           to the log file.
         :param commit_interval: actually perform a commit only every N times a commit
           is requested.
+        :param watch_interval: print watches every N seconds.
         """
 
         assert isinstance(mode, str), "mode must be a string"
@@ -496,9 +508,11 @@ class LogManager:
 
         # watch stuff
         self.watches: List[Record] = []
-        self.next_watch_tick = 1
         self.have_nonlocal_watches = False
 
+        # Interval between printing watches, in seconds
+        self.set_watch_interval(watch_interval)
+
         # database binding
         import sqlite3 as sqlite
 
@@ -516,6 +530,7 @@ class LogManager:
 
             if mode == "wu" and not file_base == ":memory:":
                 if self.is_parallel:
+                    assert self.mpi_comm
                     suffix = self.mpi_comm.bcast(_get_unique_suffix(),
                                                  root=self.head_rank)
                 else:
@@ -546,6 +561,7 @@ class LogManager:
 
                 # set globally unique run_id
                 if self.is_parallel:
+                    assert self.mpi_comm
                     self.set_constant("unique_run_id",
                             self.mpi_comm.bcast(_get_unique_id(),
                                 root=self.head_rank))
@@ -553,6 +569,7 @@ class LogManager:
                     self.set_constant("unique_run_id", _get_unique_id())
 
                 if self.is_parallel:
+                    assert self.mpi_comm
                     self.set_constant("rank_count", self.mpi_comm.Get_size())
                 else:
                     self.set_constant("rank_count", 1)
@@ -707,8 +724,20 @@ class LogManager:
 
             self.watches.append(watch_info)
 
-    def set_constant(self, name: str, value: object) -> None:
-        """Make a named, constant value available in the log."""
+    def set_watch_interval(self, interval: float) -> None:
+        """Set the interval (in seconds) between the time watches are printed.
+
+        :param interval: watch printing interval in seconds.
+        """
+        self.watch_interval = interval
+        self.next_watch_tick = self.tick_count + 1
+
+    def set_constant(self, name: str, value: Any) -> None:
+        """Make a named, constant value available in the log.
+
+        :param name: the name of the constant.
+        :param value: the value of the constant.
+        """
         existed = name in self.constants
         self.constants[name] = value
 
@@ -737,6 +766,19 @@ class LogManager:
             print("while adding datapoint for '%s':" % name)
             raise
 
+    def _update_t_log(self, name: str, value: float) -> None:
+        if value is None:
+            return
+
+        self.last_values[name] = value
+
+        try:
+            self.db_conn.execute(f"update {name} set value = {float(value)} \
+                where rank = {self.rank} and step = {self.tick_count}")
+        except Exception:
+            print("while adding datapoint for '%s':" % name)
+            raise
+
     def _gather_for_descriptor(self, gd) -> None:
         if self.tick_count % gd.interval == 0:
             q_value = gd.quantity()
@@ -793,8 +835,6 @@ class LogManager:
         for gd in self.after_gather_descriptors:
             self._gather_for_descriptor(gd)
 
-        self.tick_count += 1
-
         if tick_start_time - self.start_time > 15*60:
             save_interval = 5*60
         else:
@@ -804,11 +844,18 @@ class LogManager:
             self.save()
 
         # print watches
-        if self.tick_count == self.next_watch_tick:
+        if self.tick_count+1 >= self.next_watch_tick:
             self._watch_tick()
 
         self.t_log += time() - tick_start_time
 
+        # Adjust log update time(s), t_log
+        for gd in self.after_gather_descriptors:
+            if isinstance(gd.quantity, LogUpdateDuration):
+                self._update_t_log(gd.quantity.name, gd.quantity())
+
+        self.tick_count += 1
+
     def _commit(self) -> None:
         self.commit_countdown -= 1
         if self.commit_countdown <= 0:
@@ -826,7 +873,11 @@ class LogManager:
         self.last_save_time = time()
 
     def add_quantity(self, quantity: LogQuantity, interval: int = 1) -> None:
-        """Add an object derived from :class:`LogQuantity` to this manager."""
+        """Add a :class:`LogQuantity` to this manager.
+
+        :param quantity: add the specified :class:`LogQuantity`.
+        :param interval: interval (in time steps) when to gather this quantity.
+        """
 
         def add_internal(name, unit, description, def_agg):
             logger.debug("add log quantity '%s'" % name)
@@ -987,7 +1038,7 @@ class LogManager:
         outf.close()
 
     def plot_matplotlib(self, expr_x, expr_y) -> None:
-        from matplotlib.pyplot import xlabel, ylabel, plot  # type: ignore
+        from matplotlib.pyplot import xlabel, ylabel, plot
 
         (data_x, descr_x, unit_x), (data_y, descr_y, unit_y) = \
                 self.get_plot_data(expr_x, expr_y)
@@ -1088,6 +1139,11 @@ class LogManager:
 
         return parsed, dep_data
 
+    def _calculate_next_watch_tick(self) -> None:
+        ticks_per_interval = (self.tick_count/max(1, time()-self.start_time)
+                         * self.watch_interval)
+        self.next_watch_tick = self.tick_count + int(max(1, ticks_per_interval))
+
     def _watch_tick(self) -> None:
         """Print the watches after a tick."""
         if not self.have_nonlocal_watches and self.rank != self.head_rank:
@@ -1102,6 +1158,8 @@ class LogManager:
             gathered_data = [data_block]
 
         if self.rank == self.head_rank:
+            assert gathered_data
+
             values: Dict[str, list] = {}
             for data_block in gathered_data:
                 for name, value in data_block.items():
@@ -1123,8 +1181,7 @@ class LogManager:
                         compute_watch_str(watch) for watch in self.watches),
                       flush=True)
 
-        ticks_per_sec = self.tick_count/max(1, time()-self.start_time)
-        self.next_watch_tick = self.tick_count + int(max(1, ticks_per_sec))
+        self._calculate_next_watch_tick()
 
         if self.mpi_comm is not None and self.have_nonlocal_watches:
             self.next_watch_tick = self.mpi_comm.bcast(
@@ -1140,16 +1197,19 @@ class LogManager:
 class _SubTimer:
     def __init__(self, itimer) -> None:
         self.itimer = itimer
-        self.start_time = time()
         self.elapsed = 0
 
+    def start(self):
+        self.start_time = time()
+        return self
+
     def stop(self):
         self.elapsed += time() - self.start_time
         del self.start_time
         return self
 
     def __enter__(self) -> None:
-        pass
+        self.start()
 
     def __exit__(self, exc_type, exc_val, exc_tb) -> None:
         self.stop()
@@ -1157,7 +1217,7 @@ class _SubTimer:
 
     def submit(self) -> None:
         self.itimer.add_time(self.elapsed)
-        del self.elapsed
+        self.elapsed = 0
 
 
 class IntervalTimer(PostLogQuantity):
@@ -1165,17 +1225,23 @@ class IntervalTimer(PostLogQuantity):
     sub-timers, or by explicitly calling :meth:`add_time`.
 
     .. automethod:: __init__
+    .. automethod:: get_sub_timer
     .. automethod:: start_sub_timer
     .. automethod:: add_time
     """
 
-    def __init__(self, name: str, description: str = None) -> None:
+    def __init__(self, name: str, description: Optional[str] = None) -> None:
         LogQuantity.__init__(self, name, "s", description)
         self.elapsed: float = 0
 
-    def start_sub_timer(self):
+    def get_sub_timer(self):
         return _SubTimer(self)
 
+    def start_sub_timer(self):
+        sub_timer = _SubTimer(self)
+        sub_timer.start()
+        return sub_timer
+
     def add_time(self, t: float) -> None:
         self.start_time = time()
         self.elapsed += t
@@ -1186,14 +1252,12 @@ class IntervalTimer(PostLogQuantity):
         return result
 
 
-class LogUpdateDuration(LogQuantity):
+class LogUpdateDuration(PostLogQuantity):
     """Records how long the last log update in :class:`LogManager` took.
 
     .. automethod:: __init__
     """
 
-    # FIXME this is off by one tick
-
     def __init__(self, mgr: LogManager, name: str = "t_log") -> None:
         LogQuantity.__init__(self, name, "s", "Time spent updating the log")
         self.log_manager = mgr
@@ -1210,7 +1274,8 @@ class EventCounter(PostLogQuantity):
     .. automethod:: transfer
     """
 
-    def __init__(self, name: str = "interval", description: str = None) -> None:
+    def __init__(self, name: str = "interval",
+                 description: Optional[str] = None) -> None:
         PostLogQuantity.__init__(self, name, "1", description)
         self.events = 0
 
@@ -1317,7 +1382,7 @@ class InitTime(LogQuantity):
 
         import os
         try:
-            import psutil  # type: ignore
+            import psutil
         except ModuleNotFoundError:
             from warnings import warn
             warn("Measuring the init time requires the 'psutil' module.")
@@ -1381,6 +1446,7 @@ def add_general_quantities(mgr: LogManager) -> None:
     mgr.add_quantity(LogUpdateDuration(mgr))
     mgr.add_quantity(TimestepCounter())
     mgr.add_quantity(InitTime())
+    mgr.add_quantity(MemoryHwm())
 
 
 class SimulationTime(TimeTracker, LogQuantity):
@@ -1416,8 +1482,12 @@ def set_dt(mgr: LogManager, dt: float) -> None:
                 gd.quantity.set_dt(dt)
 
 
-def add_simulation_quantities(mgr: LogManager, dt: float = None) -> None:
-    """Add :class:`LogQuantity` objects relating to simulation time."""
+def add_simulation_quantities(mgr: LogManager, dt: Optional[float] = None) -> None:
+    """Add :class:`LogQuantity` objects relating to simulation time.
+
+    :param mgr: the :class:`LogManager` instance.
+    :param dt: (deprecated, use :meth:`set_dt` instead)
+    """
     if dt is not None:
         from warnings import warn
         warn("Specifying dt ahead of time is a deprecated practice. "
@@ -1444,6 +1514,24 @@ def add_run_info(mgr: LogManager) -> None:
     mgr.set_constant("date", strftime("%a, %d %b %Y %H:%M:%S %Z", localtime()))
     mgr.set_constant("unixtime", time())
 
+
+class MemoryHwm(PostLogQuantity):
+    """Record (monotonically increasing) memory high water mark (HWM) in MBytes."""
+    def __init__(self, name: str = "memory_usage_hwm") -> None:
+        PostLogQuantity.__init__(self, name, "MByte", "Memory High Water Mark")
+        import os
+        if os.uname().sysname == "Linux":
+            self.fac = 1024
+        elif os.uname().sysname == "Darwin":
+            self.fac = 1024*1024
+        else:
+            raise ValueError("MemoryHwm is only supported on Linux/Mac.")
+
+    def __call__(self) -> float:
+        from resource import RUSAGE_SELF, getrusage
+        res = getrusage(RUSAGE_SELF)
+        return res.ru_maxrss / self.fac
+
 # }}}
 
 # vim: foldmethod=marker
